@@ -38,6 +38,7 @@ static gboolean gconf_handle_corba_exception(CORBA_Environment* ev, GError** err
 /* just returns TRUE if there's an exception indicating the server is
    probably hosed; no side effects */
 static gboolean gconf_server_broken(CORBA_Environment* ev);
+static void gconf_detach_config_server(void);
 
 /* Maximum number of times to try re-spawning the server if it's down. */
 #define MAX_RETRIES 1
@@ -47,11 +48,19 @@ gconf_key_check(const gchar* key, GError** err)
 {
   gchar* why = NULL;
 
-  if (!gconf_valid_key(key, &why))
+  if (key == NULL)
     {
       if (err)
-        *err = gconf_error_new(GCONF_ERROR_BAD_KEY, _("`%s': %s"),
-                               key, why);
+        *err = gconf_error_new (GCONF_ERROR_BAD_KEY,
+                                _("Key \"%s\" is NULL"),
+                                key);
+      return FALSE;
+    }
+  else if (!gconf_valid_key (key, &why))
+    {
+      if (err)
+        *err = gconf_error_new (GCONF_ERROR_BAD_KEY, _("\"%s\": %s"),
+                                key, why);
       g_free(why);
       return FALSE;
     }
@@ -118,7 +127,6 @@ static ConfigServer   gconf_get_config_server    (gboolean     start_if_not_foun
 
 /* Forget our current server object reference, so the next call to
    gconf_get_config_server will have to try to respawn the server */
-static void           gconf_detach_config_server (void);
 static ConfigListener gconf_get_config_listener  (void);
 
 static void           gconf_engine_detach       (GConfEngine     *conf);
@@ -134,7 +142,7 @@ static ConfigDatabase gconf_engine_get_database (GConfEngine     *conf,
 
 #define CHECK_OWNER_USE(engine)   \
   do { if ((engine)->owner && (engine)->owner_use_count == 0) \
-     g_warning ("You can't use a GConfEngine that has an active GConfClient wrapper object. Use GConfClient API instead.");  \
+     g_warning ("%s: You can't use a GConfEngine that has an active GConfClient wrapper object. Use GConfClient API instead.", G_GNUC_FUNCTION);  \
   } while (0)
 
 static void         register_engine           (GConfEngine    *conf);
@@ -243,6 +251,20 @@ lookup_engine_by_database (ConfigDatabase db)
     return NULL;
 }
 
+static void
+database_rec_release (gpointer rec)
+{
+  GConfEngine *conf = rec;
+  CORBA_Environment ev;
+
+  CORBA_exception_init (&ev);
+
+  CORBA_Object_release (conf->database, &ev);
+  conf->database = CORBA_OBJECT_NIL;
+  
+  CORBA_exception_free (&ev);
+}
+
 /* This takes ownership of the ConfigDatabase */
 static void
 gconf_engine_set_database (GConfEngine *conf,
@@ -253,8 +275,11 @@ gconf_engine_set_database (GConfEngine *conf,
   conf->database = db;
 
   if (engines_by_db == NULL)
-    engines_by_db = g_hash_table_new ((GHashFunc) gconf_CORBA_Object_hash,
-                                      (GCompareFunc) gconf_CORBA_Object_equal);
+    engines_by_db = g_hash_table_new_full (
+	    (GHashFunc) gconf_CORBA_Object_hash,
+	    (GCompareFunc) gconf_CORBA_Object_equal,
+	    NULL,
+	    database_rec_release);
   
   g_hash_table_insert (engines_by_db, conf->database, conf);  
 }
@@ -262,16 +287,9 @@ gconf_engine_set_database (GConfEngine *conf,
 static void
 gconf_engine_detach (GConfEngine *conf)
 {
-  CORBA_Environment ev;
-
-  CORBA_exception_init (&ev);
-  
-  if (!CORBA_Object_is_nil (conf->database, &ev))
+  if (conf->database != CORBA_OBJECT_NIL)
     {
       g_hash_table_remove (engines_by_db, conf->database);
-      
-      CORBA_Object_release (conf->database, &ev);
-      conf->database = CORBA_OBJECT_NIL;
     }
 }
 
@@ -303,7 +321,7 @@ gconf_engine_connect (GConfEngine *conf,
     db = ConfigServer_get_default_database (cs, &ev);      
   else
     db = ConfigServer_get_database (cs, conf->address, &ev);
-  
+
   if (gconf_server_broken(&ev))
     {
       if (tries < MAX_RETRIES)
@@ -476,7 +494,7 @@ gconf_engine_get_for_address (const gchar* address, GError** err)
 }
 
 void
-gconf_engine_ref             (GConfEngine* conf)
+gconf_engine_ref(GConfEngine* conf)
 {
   g_return_if_fail(conf != NULL);
   g_return_if_fail(conf->refcount > 0);
@@ -485,7 +503,7 @@ gconf_engine_ref             (GConfEngine* conf)
 }
 
 void         
-gconf_engine_unref        (GConfEngine* conf)
+gconf_engine_unref(GConfEngine* conf)
 {
   g_return_if_fail(conf != NULL);
   g_return_if_fail(conf->refcount > 0);
@@ -1209,6 +1227,100 @@ gconf_engine_unset (GConfEngine* conf, const gchar* key, GError** err)
   return TRUE;
 }
 
+/**
+ * gconf_engine_recursive_unset:
+ * @engine: a #GConfEngine
+ * @key: a key or directory name
+ * @flags: change how the unset is done
+ * @err: return location for a #GError, or %NULL to ignore errors
+ * 
+ * Unsets all keys below @key, including @key itself.  If any unset
+ * fails, continues on to unset as much as it can. The first
+ * failure is returned in @err.
+ *
+ * Returns: %FALSE if error is set
+ **/
+gboolean
+gconf_engine_recursive_unset (GConfEngine    *conf,
+                              const char     *key,
+                              GConfUnsetFlags flags,
+                              GError        **err)
+{
+  CORBA_Environment ev;
+  ConfigDatabase3 db;
+  gint tries = 0;
+  ConfigDatabase3_UnsetFlags corba_flags;
+  
+  g_return_val_if_fail (conf != NULL, FALSE);
+  g_return_val_if_fail (key != NULL, FALSE);
+  g_return_val_if_fail (err == NULL || *err == NULL, FALSE);
+
+  CHECK_OWNER_USE (conf);
+  
+  if (!gconf_key_check (key, err))
+    return FALSE;
+
+  if (gconf_engine_is_local (conf))
+    {
+      GError* error = NULL;
+      
+      gconf_sources_recursive_unset (conf->local_sources, key, NULL,
+                                     flags, NULL, &error);
+
+      if (error != NULL)
+        {
+          if (err)
+            *err = error;
+          else
+            {
+              g_error_free (error);
+            }
+          return FALSE;
+        }
+      
+      return TRUE;
+    }
+
+  g_assert (!gconf_engine_is_local (conf));
+  
+  CORBA_exception_init(&ev);
+
+  corba_flags = 0;
+  if (flags & GCONF_UNSET_INCLUDING_SCHEMA_NAMES)
+    corba_flags |= ConfigDatabase3_UNSET_INCLUDING_SCHEMA_NAMES;
+  
+ RETRY:
+  
+  db = (ConfigDatabase3) gconf_engine_get_database (conf, TRUE, err);
+
+  if (db == CORBA_OBJECT_NIL)
+    {
+      g_return_val_if_fail (err == NULL || *err != NULL, FALSE);
+
+      return FALSE;
+    }
+
+  ConfigDatabase3_recursive_unset (db, key, corba_flags, &ev);
+
+  if (gconf_server_broken (&ev))
+    {
+      if (tries < MAX_RETRIES)
+        {
+          ++tries;
+          CORBA_exception_free(&ev);
+          gconf_engine_detach(conf);
+          goto RETRY;
+        }
+    }
+  
+  if (gconf_handle_corba_exception (&ev, err))
+    return FALSE;
+
+  g_return_val_if_fail (err == NULL || *err == NULL, FALSE);
+  
+  return TRUE;
+}
+
 gboolean
 gconf_engine_associate_schema  (GConfEngine* conf, const gchar* key,
                                 const gchar* schema_key, GError** err)
@@ -1217,24 +1329,21 @@ gconf_engine_associate_schema  (GConfEngine* conf, const gchar* key,
   ConfigDatabase db;
   gint tries = 0;
 
-  g_return_val_if_fail(conf != NULL, FALSE);
-  g_return_val_if_fail(key != NULL, FALSE);
-  g_return_val_if_fail(schema_key != NULL, FALSE);
-  g_return_val_if_fail(err == NULL || *err == NULL, FALSE);
-
-  CHECK_OWNER_USE (conf);
+  g_return_val_if_fail (conf != NULL, FALSE);
+  g_return_val_if_fail (key != NULL, FALSE);
+  g_return_val_if_fail (err == NULL || *err == NULL, FALSE);
   
-  if (!gconf_key_check(key, err))
+  if (!gconf_key_check (key, err))
     return FALSE;
 
-  if (!gconf_key_check(schema_key, err))
+  if (schema_key && !gconf_key_check (schema_key, err))
     return FALSE;
 
   if (gconf_engine_is_local(conf))
     {
       GError* error = NULL;
       
-      gconf_sources_set_schema(conf->local_sources, key, schema_key, &error);
+      gconf_sources_set_schema (conf->local_sources, key, schema_key, &error);
 
       if (error != NULL)
         {
@@ -1250,32 +1359,33 @@ gconf_engine_associate_schema  (GConfEngine* conf, const gchar* key,
       return TRUE;
     }
 
-  g_assert(!gconf_engine_is_local(conf));
+  g_assert (!gconf_engine_is_local (conf));
   
-  CORBA_exception_init(&ev);
+  CORBA_exception_init (&ev);
 
  RETRY:
   
-  db = gconf_engine_get_database(conf, TRUE, err);
+  db = gconf_engine_get_database (conf, TRUE, err);
 
   if (db == CORBA_OBJECT_NIL)
     {
-      g_return_val_if_fail(err == NULL || *err != NULL, FALSE);
+      g_return_val_if_fail (err == NULL || *err != NULL, FALSE);
 
       return FALSE;
     }
 
-  ConfigDatabase_set_schema(db,
-                            (gchar*)key,
-                            (gchar*)schema_key,
-                            &ev);
+  ConfigDatabase_set_schema (db,
+                             key,
+                             /* empty string means unset */
+                             schema_key ? schema_key : "",
+                             &ev);
 
-  if (gconf_server_broken(&ev))
+  if (gconf_server_broken (&ev))
     {
       if (tries < MAX_RETRIES)
         {
           ++tries;
-          CORBA_exception_free(&ev);
+          CORBA_exception_free (&ev);
           gconf_engine_detach (conf);
           goto RETRY;
         }
@@ -1284,7 +1394,7 @@ gconf_engine_associate_schema  (GConfEngine* conf, const gchar* key,
   if (gconf_handle_corba_exception(&ev, err))
     return FALSE;
 
-  g_return_val_if_fail(*err == NULL, FALSE);
+  g_return_val_if_fail (*err == NULL, FALSE);
   
   return TRUE;
 }
@@ -1827,7 +1937,8 @@ gconf_engine_remove_dir (GConfEngine* conf,
   g_return_if_fail(dir != NULL);
   g_return_if_fail(err == NULL || *err == NULL);
 
-  CHECK_OWNER_USE (conf);
+  /* FIXME we have no GConfClient method for doing this */
+  /*   CHECK_OWNER_USE (conf); */
   
   if (!gconf_key_check(dir, err))
     return;
@@ -1962,13 +2073,14 @@ try_to_contact_server (gboolean start_if_not_found,
       
       if (ev._major != CORBA_NO_EXCEPTION)
 	{
-	  server = CORBA_OBJECT_NIL;
           g_set_error (err,
                        GCONF_ERROR,
                        GCONF_ERROR_NO_SERVER,
                        _("Adding client to server's list failed, CORBA error: %s"),
                        CORBA_exception_id (&ev));
 
+	  CORBA_Object_release (server, &ev);
+	  server = CORBA_OBJECT_NIL;
           CORBA_exception_free(&ev);
 	}
     }
@@ -1996,29 +2108,58 @@ gconf_get_config_server(gboolean start_if_not_found, GError** err)
   return server; /* return what we have, NIL or not */
 }
 
-static void
+ConfigListener listener = CORBA_OBJECT_NIL;
+
+void
 gconf_detach_config_server(void)
 {  
+  CORBA_Environment ev;
+
+  CORBA_exception_init(&ev);
+
+  if (listener != CORBA_OBJECT_NIL)
+    {
+      CORBA_Object_release(listener, &ev);
+      listener = CORBA_OBJECT_NIL;
+    }
+
   if (server != CORBA_OBJECT_NIL)
     {
-      CORBA_Environment ev;
-      
-      CORBA_exception_init(&ev);
-
       CORBA_Object_release(server, &ev);
 
       if (ev._major != CORBA_NO_EXCEPTION)
         {
           g_warning("Exception releasing gconfd server object: %s",
                     CORBA_exception_id(&ev));
-          CORBA_exception_free(&ev);
         }
 
       server = CORBA_OBJECT_NIL;
     }
+
+  CORBA_exception_free(&ev);
+
+  if (engines_by_db != NULL)
+    {
+      g_hash_table_destroy (engines_by_db);
+      engines_by_db = NULL;
+    }
+}
+
+/**
+ * gconf_debug_shutdown:
+ * @void: 
+ * 
+ * Detach from the config server and release
+ * all related resources
+ **/
+int
+gconf_debug_shutdown (void)
+{
+  gconf_detach_config_server ();
+
+  return gconf_orb_release ();
 }
 
-ConfigListener listener = CORBA_OBJECT_NIL;
 static void notify                  (PortableServer_Servant     servant,
                                      ConfigDatabase             db,
                                      CORBA_unsigned_long        cnxn,
@@ -2196,13 +2337,12 @@ drop_all_caches (PortableServer_Servant     _servant,
 
 static ConfigListener 
 gconf_get_config_listener(void)
-{
-  static ConfigListener listener = CORBA_OBJECT_NIL;
-  
+{  
   if (listener == CORBA_OBJECT_NIL)
     {
       CORBA_Environment ev;
       PortableServer_POA poa;
+      PortableServer_POAManager poa_mgr;
 
       CORBA_exception_init (&ev);
       POA_ConfigListener__init (&poa_listener_servant, &ev);
@@ -2215,7 +2355,8 @@ gconf_get_config_listener(void)
 
       g_assert (ev._major == CORBA_NO_EXCEPTION);
 
-      PortableServer_POAManager_activate(PortableServer_POA__get_the_POAManager(poa, &ev), &ev);
+      poa_mgr = PortableServer_POA__get_the_POAManager (poa, &ev);
+      PortableServer_POAManager_activate (poa_mgr, &ev);
 
       g_assert (ev._major == CORBA_NO_EXCEPTION);
 
@@ -2223,6 +2364,9 @@ gconf_get_config_listener(void)
                                                          &poa_listener_servant,
                                                          &ev);
 
+      CORBA_Object_release ((CORBA_Object) poa_mgr, &ev);
+      CORBA_Object_release ((CORBA_Object) poa, &ev);
+
       g_assert (listener != CORBA_OBJECT_NIL);
       g_assert (ev._major == CORBA_NO_EXCEPTION);
     }
@@ -2480,8 +2624,13 @@ gconf_unescape_key (const char *escaped_key,
 gboolean
 gconf_key_is_below   (const gchar* above, const gchar* below)
 {
-  int len = strlen(above);
-  if (strncmp(below, above, len) == 0)
+  int len;
+
+  if (above[0] == '/' && above[1] == '\0')
+    return TRUE;
+  
+  len = strlen (above);
+  if (strncmp (below, above, len) == 0)
     {
       /* only if this is a complete key component,
        * so that /foo is not above /foofoo/bar */
@@ -3346,4 +3495,3 @@ gconf_enum_to_string (GConfEnumStringPair lookup_table[],
 
   return NULL;
 }
-
