@@ -107,6 +107,8 @@ static GConfDatabase*       obtain_database (const gchar *address,
 static void                 drop_old_databases (void);
 static gboolean             no_databases_in_use (void);
 
+static void gconf_handle_segv (int signum);
+
 /*
  * Flag indicating that we are shutting down, so return errors
  * on any attempted operation. We do this instead of unregistering with
@@ -358,15 +360,17 @@ signal_handler (int signo)
   
   ++in_fatal;
   
-  switch(signo) {
+  switch (signo) {
     /* Fast cleanup only */
   case SIGSEGV:
   case SIGBUS:
   case SIGILL:
     enter_shutdown ();
-    gconf_log(GCL_ERR,
-              _("Received signal %d, dumping core. Please report a GConf bug."),
-              signo);
+    gconf_log (GCL_ERR,
+               _("Received signal %d, dumping core. Please report a GConf bug."),
+               signo);
+    if (g_getenv ("DISPLAY"))
+      gconf_handle_segv (signo);
     abort ();
     break;
 
@@ -428,7 +432,7 @@ log_handler (const gchar   *log_domain,
 {
   GConfLogPriority pri = GCL_WARNING;
   
-  switch (log_level)
+  switch (log_level & G_LOG_LEVEL_MASK)
     {
     case G_LOG_LEVEL_ERROR:
     case G_LOG_LEVEL_CRITICAL:
@@ -460,7 +464,6 @@ main(int argc, char** argv)
 {
   struct sigaction act;
   sigset_t empty_mask;
-  PortableServer_ObjectId* objid;
   CORBA_Environment ev;
   CORBA_ORB orb;
   gchar* logname;
@@ -500,6 +503,10 @@ main(int argc, char** argv)
 	  dup2 (dev_null_fd, 2);
 	}
     }
+  else
+    {
+      gconf_log_debug_messages = TRUE;
+    }
   
   umask (022);
   
@@ -515,6 +522,12 @@ main(int argc, char** argv)
 
   g_log_set_handler (NULL, G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
                      log_handler, NULL);
+
+  g_log_set_handler ("GLib", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
+                     log_handler, NULL);
+
+  g_log_set_handler ("GLib-GObject", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
+                     log_handler, NULL);
   
   /* openlog() does not copy logname - what total brokenness.
      So we free it at the end of main() */
@@ -553,8 +566,6 @@ main(int argc, char** argv)
   the_poa = (PortableServer_POA)CORBA_ORB_resolve_initial_references(orb, "RootPOA", &ev);
   PortableServer_POAManager_activate(PortableServer_POA__get_the_POAManager(the_poa, &ev), &ev);
 
-  objid = PortableServer_POA_activate_object(the_poa, &poa_server_servant, &ev);
-  
   server = PortableServer_POA_servant_to_reference(the_poa,
                                                    &poa_server_servant,
                                                    &ev);
@@ -655,8 +666,9 @@ main(int argc, char** argv)
   daemon_lock = NULL;
   
   gconf_log (GCL_INFO, _("Exiting"));
-  
-  g_free (logname);
+
+  /* Can't do this due to stupid atexit() handler that calls g_log stuff */
+  /*   g_free (logname); */
   
   return exit_code;
 }
@@ -1028,11 +1040,21 @@ gconf_set_exception(GError** error,
       case GCONF_ERROR_IN_SHUTDOWN:
         ce->err_no = ConfigInShutdown;
         break;
-        
+      case GCONF_ERROR_OVERRIDDEN:
+        ce->err_no = ConfigOverridden;
+        break;
+      case GCONF_ERROR_LOCK_FAILED:
+        ce->err_no = ConfigLockFailed;
+        break;
+
+      case GCONF_ERROR_OAF_ERROR:
+      case GCONF_ERROR_LOCAL_ENGINE:
       case GCONF_ERROR_NO_SERVER:
       case GCONF_ERROR_SUCCESS:
       default:
+        gconf_log (GCL_ERR, "Unhandled error code %d", en);
         g_assert_not_reached();
+        break;
       }
 
     CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
@@ -1700,11 +1722,11 @@ restore_listener (GConfDatabase* db,
                  "Failed to cancel previous daemon's listener in saved state file: %s",
                  err->message);
       g_error_free (err);
-    }  
+    }
+  
+  new_cnxn = gconf_database_readd_listener (db, cl, "from-saved-state", lle->location);
 
-  new_cnxn = gconf_database_readd_listener (db, cl, lle->location);
-
-  gconf_log (GCL_DEBUG, "Attempting to update listener from saved state file, old connection %u, new connectin %u", (guint) lle->connection_id, (guint) new_cnxn);
+  gconf_log (GCL_DEBUG, "Attempting to update listener from saved state file, old connection %u, new connection %u", (guint) lle->connection_id, (guint) new_cnxn);
   
   ConfigListener_update_listener (cl,
                                   db->objref,
@@ -2224,22 +2246,31 @@ drop_old_clients (void)
       while (tmp != NULL)
         {
           ConfigListener cl = tmp->data;
-          
-          ConfigListener_ping (cl, &ev);
+          CORBA_boolean result;
 
+          result = CORBA_Object_non_existent (cl, &ev);
+          
           if (ev._major != CORBA_NO_EXCEPTION)
             {
-              gconf_log (GCL_DEBUG, "Removing stale client");
+              gconf_log (GCL_WARNING, "Exception from CORBA_Object_non_existent(), assuming stale listener");
+              CORBA_exception_free (&ev);
+              CORBA_exception_init (&ev);
+              result = TRUE;
+            }
+
+          if (result)
+            {
+              gconf_log (GCL_DEBUG, "removing stale client in drop_old_clients");
               
               remove_client (cl);
-              
-              CORBA_exception_free (&ev);
             }
           
           tmp = g_slist_next (tmp);
         }
 
       g_slist_free (clients);
+
+      CORBA_exception_free (&ev);
     }
 }
 
@@ -2251,3 +2282,76 @@ client_count (void)
   else
     return g_hash_table_size (client_table);
 }
+
+
+static void
+gconf_handle_segv (int signum)
+{
+  static int in_segv = 0;
+  pid_t pid;
+	
+  in_segv++;
+
+  if (in_segv > 2)
+    {
+      /* The fprintf() was segfaulting, we are just totally hosed */
+      _exit(1);
+    }
+  else if (in_segv > 1)
+    {
+      /* dialog display isn't working out */
+      fprintf (stderr, _("Multiple segmentation faults occurred; can't display error dialog\n"));
+      _exit(1);
+    }
+        
+  pid = fork();
+
+  if (pid < 0)
+    {
+      /* Eeeek! Can't show dialog */
+      fprintf (stderr, _("Segmentation fault!\n"
+                         "Cannot display crash dialog\n"));
+      
+      /* Don't use app attributes here - a lot of things are probably hosed */
+      if (g_getenv ("GNOME_DUMP_CORE"))
+        abort ();
+      
+      _exit (1);
+    }
+  else if (pid > 0)
+    {
+      /* Wait for user to see the dialog, then exit. */
+      /* Why wait at all? Because we want to allow people to attach to the
+         process */
+      int estatus;
+      pid_t eret;
+      
+      eret = waitpid (pid, &estatus, 0);
+      
+      /* Don't use app attributes here - a lot of things are probably hosed */
+      if(g_getenv ("GNOME_DUMP_CORE"))
+        abort ();
+      
+      _exit (1);
+    }
+  else /* pid == 0 */
+    {
+      char buf[32];
+      
+      g_snprintf (buf, sizeof (buf), "%d", signum);
+      
+      /* Child process */
+      execl (GCONF_BINDIR "/gnome_segv2", GCONF_BINDIR "/gnome_segv",
+             GCONFD, buf,
+             VERSION, NULL);
+      
+      execlp ("gnome_segv2", "gnome_segv2",
+              GCONFD, buf,
+              GCONFD, NULL);
+      
+      _exit (99);
+    }
+  
+  in_segv--;
+}
+
